@@ -7,6 +7,7 @@ __all__ = ['fetch']
 
 import errno
 import functools
+import glob
 import io
 import itertools
 import json
@@ -27,7 +28,6 @@ except ImportError:
 
 import portage
 portage.proxy.lazyimport.lazyimport(globals(),
-	'glob:glob',
 	'portage.package.ebuild.config:check_config_instance,config',
 	'portage.package.ebuild.doebuild:doebuild_environment,' + \
 		'_doebuild_spawn',
@@ -272,7 +272,13 @@ class FlatLayout(object):
 	def get_filenames(self, distdir):
 		for dirpath, dirnames, filenames in os.walk(distdir,
 				onerror=_raise_exc):
-			return iter(filenames)
+			for filename in filenames:
+				try:
+					yield portage._unicode_decode(filename, errors='strict')
+				except UnicodeDecodeError:
+					# Ignore it. Distfiles names must have valid UTF8 encoding.
+					pass
+			return
 
 	@staticmethod
 	def verify_args(args):
@@ -301,8 +307,12 @@ class FilenameHashLayout(object):
 			c = c // 4
 			pattern += c * '[0-9a-f]' + '/'
 		pattern += '*'
-		return (x.rsplit('/', 1)[1]
-				for x in glob(os.path.join(distdir, pattern)))
+		for x in glob.iglob(portage._unicode_encode(os.path.join(distdir, pattern), errors='strict')):
+			try:
+				yield portage._unicode_decode(x, errors='strict').rsplit('/', 1)[1]
+			except UnicodeDecodeError:
+				# Ignore it. Distfiles names must have valid UTF8 encoding.
+				pass
 
 	@staticmethod
 	def verify_args(args):
@@ -382,7 +392,7 @@ class MirrorLayoutConfig(object):
 		return ret
 
 
-def get_mirror_url(mirror_url, filename, cache_path=None):
+def get_mirror_url(mirror_url, filename, mysettings, cache_path=None):
 	"""
 	Get correct fetch URL for a given file, accounting for mirror
 	layout configuration.
@@ -408,23 +418,22 @@ def get_mirror_url(mirror_url, filename, cache_path=None):
 	if ts >= time.time() - 86400:
 		mirror_conf.deserialize(data)
 	else:
+		tmpfile = '.layout.conf.%s' % urlparse(mirror_url).hostname
 		try:
-			f = urlopen(mirror_url + '/distfiles/layout.conf')
-			try:
-				data = io.StringIO(f.read().decode('utf8'))
-			finally:
-				f.close()
-
-			mirror_conf.read_from_file(data)
+			if fetch({tmpfile: (mirror_url + '/distfiles/layout.conf',)},
+					mysettings, force=1, try_mirrors=0):
+				tmpfile = os.path.join(mysettings['DISTDIR'], tmpfile)
+				mirror_conf.read_from_file(tmpfile)
+			else:
+				raise IOError()
 		except (ConfigParserError, IOError, UnicodeDecodeError):
-			# Do not cache negative results.
-			cache_path = None
-
-		cache[mirror_url] = (time.time(), mirror_conf.serialize())
-		if cache_path is not None:
-			f = atomic_ofstream(cache_path, 'w')
-			json.dump(cache, f)
-			f.close()
+			pass
+		else:
+			cache[mirror_url] = (time.time(), mirror_conf.serialize())
+			if cache_path is not None:
+				f = atomic_ofstream(cache_path, 'w')
+				json.dump(cache, f)
+				f.close()
 
 	return (mirror_url + "/distfiles/" +
 			mirror_conf.get_best_supported_layout().get_path(filename))
@@ -432,8 +441,45 @@ def get_mirror_url(mirror_url, filename, cache_path=None):
 
 def fetch(myuris, mysettings, listonly=0, fetchonly=0,
 	locks_in_subdir=".locks", use_locks=1, try_mirrors=1, digests=None,
-	allow_missing_digests=True):
-	"fetch files.  Will use digest file if available."
+	allow_missing_digests=True, force=False):
+	"""
+	Fetch files to DISTDIR and also verify digests if they are available.
+
+	@param myuris: Maps each file name to a tuple of available fetch URIs.
+	@type myuris: dict
+	@param mysettings: Portage config instance.
+	@type mysettings: portage.config
+	@param listonly: Only print URIs and do not actually fetch them.
+	@type listonly: bool
+	@param fetchonly: Do not block for files that are locked by a
+		concurrent fetcher process. This means that the function can
+		return successfully *before* all files have been successfully
+		fetched!
+	@type fetchonly: bool
+	@param use_locks: Enable locks. This parameter is ineffective if
+		FEATURES=distlocks is disabled in the portage config!
+	@type use_locks: bool
+	@param digests: Maps each file name to a dict of digest types and values.
+	@type digests: dict
+	@param allow_missing_digests: Enable fetch even if there are no digests
+		available for verification.
+	@type allow_missing_digests: bool
+	@param force: Force download, even when a file already exists in
+		DISTDIR. This is most useful when there are no digests available,
+		since otherwise download will be automatically forced if the
+		existing file does not match the available digests. Also, this
+		avoids the need to remove the existing file in advance, which
+		makes it possible to atomically replace the file and avoid
+		interference with concurrent processes.
+	@type force: bool
+	@rtype: int
+	@return: 1 if successful, 0 otherwise.
+	"""
+
+	if force and digests:
+		# Since the force parameter can trigger unnecessary fetch when the
+		# digests match, do not allow force=True when digests are provided.
+		raise PortageException(_('fetch: force=True is not allowed when digests are provided'))
 
 	if not myuris:
 		return 1
@@ -619,7 +665,7 @@ def fetch(myuris, mysettings, listonly=0, fetchonly=0,
 				mirror_cache = None
 			for l in locations:
 				filedict[myfile].append(functools.partial(
-					get_mirror_url, l, myfile, mirror_cache))
+					get_mirror_url, l, myfile, mysettings, mirror_cache))
 		if myuri is None:
 			continue
 		if myuri[:9]=="mirror://":
@@ -854,7 +900,7 @@ def fetch(myuris, mysettings, listonly=0, fetchonly=0,
 				eout.quiet = mysettings.get("PORTAGE_QUIET") == "1"
 				match, mystat = _check_distfile(
 					myfile_path, pruned_digests, eout, hash_filter=hash_filter)
-				if match:
+				if match and not force:
 					# Skip permission adjustment for symlinks, since we don't
 					# want to modify anything outside of the primary DISTDIR,
 					# and symlinks typically point to PORTAGE_RO_DISTDIRS.
@@ -1018,10 +1064,11 @@ def fetch(myuris, mysettings, listonly=0, fetchonly=0,
 								os.unlink(download_path)
 							except EnvironmentError:
 								pass
-					elif myfile not in mydigests:
+					elif not orig_digests:
 						# We don't have a digest, but the file exists.  We must
 						# assume that it is fully downloaded.
-						continue
+						if not force:
+							continue
 					else:
 						if (mydigests[myfile].get("size") is not None
 								and mystat.st_size < mydigests[myfile]["size"]
